"""Visual search pipeline.
"""

import os
import time
import sys

import cPickle as pickle
import numpy as np

from sklearn.preprocessing import normalize

from jinja2 import Environment, FileSystemLoader

from owl.data import common
from owl.data import img_tools
from owl.data import mongo_manager
from owl.features import obj_feat_extractor
from owl.processors import dim_reductor
from owl.search import common as search_common
from owl.search import simple_matchers
from owl.pipelines.common import PipelineMode


class OneFeatVisualSearcher(object):
  """A single feature based visual search object.

  Used for both development and deployment.

  1) read in files and organize into database and query.
  2) compute features and save to database:
  [img_fn, label, type, feat1, feat2, ...].
  3) reduce feature dimension if too high (x00~1000+) and save to database.
  4) run matching.
  5) evalute.
  All db involved: [task_name]_db.raw_feats, [task_name]_db.compressed_feats,
                   [task_name]_query.raw_feats, [task_name]_query.compressed_feats.

    Attributes:
      res_dir: result directory.
  """

  res_dir = None
  logger = None
  task_name = ""
  mode = PipelineMode.DEVELOPMENT
  # db name for database.
  task_db_db_name = ""
  # db name for query.
  task_query_db_name = ""
  target_feat_name = ""
  target_feat_dim = 0
  # list of items.
  db_items = []
  query_items = []
  # {feat_name: feat_matrix}
  db_feats = {}
  query_feats = {}
  # ranked list of db ids for each query.
  ranked_lists = None
  # feature extraction.
  feat_extractor = None
  feat_compute_batch_size = 10
  # feature dim reduction.
  use_feat_compressor = False
  feat_compress_batch_size = 10
  compress_feat_dim = 128
  train_feat_compressor = False
  feat_compressor = None
  # matching.
  matcher = None
  match_protocol = search_common.MatchProtocol.APPROXIMATE
  num_trees = 100

  def __init__(self, task_name, res_dir):
    self.task_name = task_name
    self.task_db_db_name = "{}_db".format(self.task_name)
    self.task_query_db_name = "{}_query".format(self.task_name)
    self.res_dir = res_dir
    log_fn = os.path.join(res_dir, "log.txt")
    self.logger = common.get_logger("visualsearchlogger", log_fn)

  def clean_db(self):
    """Remove all related database.
    """
    db_man = mongo_manager.MongoManager()
    db_man.connect()
    db_man.delete_db(self.task_db_db_name)
    db_man.delete_db(self.task_query_db_name)

  def load_items_from_db(self):
    """Load data from local db.
    """
    # load db items.
    self.db_items = []
    item_db = mongo_manager.MongoManager()
    item_db.connect(db_name=self.task_db_db_name, collection_name="items")
    db_cursor = item_db.get_scan_cursor()
    for db_item in db_cursor:
      self.db_items.append(db_item)
    print "{} db items loaded from db.".format(len(self.db_items))
    if self.mode == PipelineMode.DEVELOPMENT:
      # load query items.
      self.query_items = []
      item_db.switch_db_collection(self.task_query_db_name, "items")
      query_cursor = item_db.get_scan_cursor()
      for query_item in query_cursor:
        self.query_items.append(query_item)
      print "{} query items loaded from db.".format(len(self.query_items))
    item_db.disconnect()

  def load_feats_from_db(self, raw_feat=False):
    """Load features from db.

    At the same time, load item with the same order.
    """

    def load_feats_items_from_db(item_db, feat_db):
      """load features and items from db.

      Args:
        item_db: item db object.
        feat_db: feat db object.
      Returns:
        items: list of items
        feats: feature ndarray.
      """
      items = []
      db_cursor = feat_db.get_scan_cursor()
      total_item_count = db_cursor.count()
      feats = np.zeros((total_item_count, self.target_feat_dim))
      for db_id, db_feat_item in enumerate(db_cursor):
        db_item_id = db_feat_item["_id"]
        db_feat = np.array(db_feat_item[self.target_feat_name]).reshape(1, -1)
        feats[db_id] = db_feat
        cur_item = item_db.query(value_list=[db_item_id])[0]
        items.append(cur_item)
        if len(items) % 1000 == 0:
          print "loaded {}/{} items".format(len(items), total_item_count)
      return items, feats

    self.logger.info("loading features from db...")
    # load db data.
    item_db = mongo_manager.MongoManager()
    item_db.connect(db_name=self.task_db_db_name, collection_name="items")
    feat_db = mongo_manager.MongoManager()
    if not raw_feat:
      feat_db.connect(
          db_name=self.task_db_db_name, collection_name="compressed_feats")
    else:
      feat_db.connect(
          db_name=self.task_db_db_name, collection_name="raw_feats")
    self.db_items, self.db_feats = load_feats_items_from_db(item_db, feat_db)

    if self.mode == PipelineMode.DEVELOPMENT:
      # load query data.
      item_db.switch_db_collection(
          db_name=self.task_query_db_name, collection_name="items")
      if not raw_feat:
        feat_db.switch_db_collection(
            db_name=self.task_query_db_name,
            collection_name="compressed_feats")
      else:
        feat_db.switch_db_collection(
            db_name=self.task_query_db_name, collection_name="raw_feats")
      self.query_items, self.query_feats = load_feats_items_from_db(item_db,
                                                                    feat_db)

    item_db.disconnect()
    feat_db.disconnect()
    self.logger.info("features loaded from db")

  def read_data_from_files(self, img_folder, img_exts):
    """Input image data.

    Args:
      img_folder: folder contains image files.
    If has subfolders, each is treated as a category with folder name.
    """
    # read all image files.
    img_fns, img_labels, _ = common.list_img_files(img_folder, img_exts)
    # split into query and database.
    db_ids, _, query_ids = common.split_train_val_test(
        img_labels, train_ratio=0.8, test_ratio=0.2)
    self.logger.info("{} db files and {} query files found.".format(
        len(db_ids), len(query_ids)))
    # save db item info.
    item_db = mongo_manager.MongoManager()
    item_db.connect(db_name=self.task_db_db_name, collection_name="items")
    item_db.delete_collection()
    for i, db_id in enumerate(db_ids):
      cur_db_item = {"img_fn": img_fns[db_id], "label": img_labels[db_id]}
      item_db.add(cur_db_item)
      if i % 100 == 0:
        self.logger.info("{}/{} db items added.".format(i, len(db_ids)))
    self.logger.info("all db items added to db.")
    # save query item info.
    item_db.switch_db_collection(self.task_query_db_name, "items")
    item_db.delete_collection()
    for i, query_id in enumerate(query_ids):
      cur_query_item = {
          "img_fn": img_fns[query_id],
          "label": img_labels[query_id]
      }
      item_db.add(cur_query_item)
      if i % 100 == 0:
        self.logger.info("{}/{} query items added.".format(i, len(query_ids)))
    self.logger.info("all query items added to db.")
    item_db.disconnect()

  def compute_feats(self):
    """[DEV] Compute features for all images.

    Args:
      features: feature extractors.
      batch_size: image batch to process.
    """

    def compute_raw_feats_(raw_feat_db, items):
      """compute raw features for a specific db.

      Args:
        raw_feat_db: db object for raw features.
        items: list of items.
      """
      startt = time.time()
      # loop over items.
      for idx in range(0, len(items), self.feat_compute_batch_size):
        try:
          batch_items = items[idx:idx + self.feat_compute_batch_size]
          rgb_imgs = []
          # group images.
          for item in batch_items:
            img_bin = img_tools.read_img_bin(item["img_fn"])
            img_arr = img_tools.img_bin_to_img_arr(img_bin)
            rgb_imgs.append(img_arr)
          # compute batch features.
          startt2 = time.time()
          batch_feats = self.feat_extractor.compute_batch(rgb_imgs)
          self.logger.info("batch feature extraction time: {}s".format(
              time.time() - startt2))
          # save feature to database.
          batch_feat_objs = []
          for item_id, item in enumerate(batch_items):
            cur_feat_obj = {}
            cur_feat_obj["_id"] = item["_id"]
            cur_feat_obj[self.target_feat_name] = batch_feats[item_id].tolist()
            batch_feat_objs.append(cur_feat_obj)
          raw_feat_db.add_many(batch_feat_objs)
          self.logger.info("database feature: {}/{}".format(
              min(idx + self.feat_compute_batch_size, len(items)), len(items)))
        except Exception as ex:
          self.logger.error("{}, skip one batch".format(
              common.get_detailed_error_msg(ex.message, sys.exc_info())))
          continue

      self.logger.info(
          "database feature extraction done. total time cost: {}s".format(
              time.time() - startt))

    # compute features for db.
    raw_feat_db = mongo_manager.MongoManager()
    raw_feat_db.connect(
        db_name=self.task_db_db_name, collection_name="raw_feats")
    raw_feat_db.delete_collection()
    compute_raw_feats_(raw_feat_db, self.db_items)

    # compute features for query.
    raw_feat_db.switch_db_collection("{}_query".format(self.task_name),
                                     "raw_feats")
    raw_feat_db.delete_collection()
    compute_raw_feats_(raw_feat_db, self.query_items)
    raw_feat_db.disconnect()

  def compress_feats(self):
    """[DEV] Compress features.
    """
    raw_feat_db = mongo_manager.MongoManager()
    raw_feat_db.connect(
        db_name=self.task_db_db_name, collection_name="raw_feats")
    if self.train_feat_compressor:
      db_item_cursor = raw_feat_db.get_scan_cursor()
      total_db_item_count = db_item_cursor.count()
      train_batch_size = min(self.feat_compress_batch_size,
                             total_db_item_count)
      self.logger.info("total db item to use for training compressor: {}".
                       format(total_db_item_count))
      db_raw_feats_batch = None
      batch_item_count = 0
      total_item_count = 0
      for db_item in db_item_cursor:
        cur_feat = np.array(db_item[self.target_feat_name])
        cur_feat = cur_feat.reshape(1, -1)
        batch_item_count += 1
        total_item_count += 1
        if db_raw_feats_batch is None:
          db_raw_feats_batch = cur_feat
        else:
          db_raw_feats_batch = np.vstack((db_raw_feats_batch, cur_feat))
        if batch_item_count >= train_batch_size:
          print "{}/{} db features loaded.".format(total_item_count,
                                                   total_db_item_count)
          print db_raw_feats_batch.shape
          self.feat_compressor.train(db_raw_feats_batch, False)
          print "dim reducer updated."
          db_raw_feats_batch = None
          batch_item_count = 0
    else:
      self.feat_compressor.save_load_model(to_save=False)

    def _compress_feats(raw_feat_db, compressed_feat_db):
      """Actually compress feats in a db.

      Args:
        raw_feat_db: raw feature db object.
        compressed_feat_db: compressed feat db object.
      """
      raw_item_cursor = raw_feat_db.get_scan_cursor()
      db_item_count = raw_item_cursor.count()
      cur_batch_size = min(self.feat_compress_batch_size, db_item_count)
      batch_item_count = 0
      processed_item_count = 0
      batch_feats = None
      batch_objs = []
      for item in raw_item_cursor:
        batch_item_count += 1
        processed_item_count += 1
        old_feat = np.array(item[self.target_feat_name]).reshape(1, -1)
        if batch_feats is None:
          batch_feats = old_feat
        else:
          batch_feats = np.vstack((batch_feats, old_feat))
        cur_obj = {"_id": item["_id"]}
        batch_objs.append(cur_obj)
        if batch_item_count >= cur_batch_size:
          new_feats = self.feat_compressor.transform(batch_feats)
          new_feats = normalize(new_feats, norm="l2")
          for idx, new_feat in enumerate(new_feats):
            batch_objs[idx][self.target_feat_name] = new_feat.tolist()
          # add to database.
          # TODO(jiefeng): use update to save storage?
          compressed_feat_db.add_many(batch_objs)
          self.logger.info("{}/{} compressed feature added to db.".format(
              processed_item_count, db_item_count))
          # clean.
          batch_item_count = 0
          batch_objs = []
          batch_feats = None
      self.logger.info("feature compression done.")

    # compress db features.
    self.logger.info("start compressing db features")
    raw_feat_db.switch_db_collection(
        db_name=self.task_db_db_name, collection_name="raw_feats")
    compressed_feat_db = mongo_manager.MongoManager()
    compressed_feat_db.connect(
        db_name=self.task_db_db_name, collection_name="compressed_feats")
    compressed_feat_db.delete_collection()
    _compress_feats(raw_feat_db, compressed_feat_db)
    # compress query features.
    self.logger.info("start compressing query features")
    raw_feat_db.switch_db_collection(
        db_name=self.task_query_db_name, collection_name="raw_feats")
    compressed_feat_db.switch_db_collection(
        db_name=self.task_query_db_name, collection_name="compressed_feats")
    compressed_feat_db.delete_collection()
    _compress_feats(raw_feat_db, compressed_feat_db)

    # clean.
    raw_feat_db.disconnect()
    compressed_feat_db.disconnect()

  def run_match(self):
    """[DEV] Perform matching between query and database.
    """
    self.logger.info("start matching...")
    # set database.
    if len(self.db_feats) == 0 or len(self.query_feats) == 0:
      self.load_feats_from_db(not self.use_feat_compressor)
    assert len(self.db_feats) > 0 and len(self.query_feats) > 0, \
      "db or query feature is empty"
    db_items = {"metadata": self.db_items}
    db_items["features"] = self.db_feats
    self.matcher.prepare_database(db_items)
    query_items = {"metadata": self.query_items}
    query_items["features"] = self.query_feats
    startt = time.time()
    self.ranked_lists = self.matcher.match(query_items, top_nn=20)
    self.logger.info("match time: {}s".format(time.time() - startt))
    self.logger.info("matching done.")

  def evaluate(self):
    """[DEV] Compute retrieval accuracy and visualize results.

    Args:
      res_dir: directory to put result data.
    """
    # get result display name from save directory.
    res_name = self.res_dir.rstrip("/")
    res_name = os.path.basename(res_name)
    self.logger.info("start evaluation...")
    # compute accuracy: top1, top5, top10, top20.
    accu_rank = [1, 5, 10, 20]
    top_k_accu = [0, 0, 0, 0]
    # for each query.
    for query_id, cur_ranked_ids in enumerate(self.ranked_lists):
      for k, rank in enumerate(accu_rank):
        top_k_accu[k] += np.mean([
            1
            if self.query_items[query_id]["label"] == self.db_items[x]["label"]
            else 0 for x in cur_ranked_ids[:rank]
        ])
    top_k_accu = [x / len(self.query_items) for x in top_k_accu]
    # visualize results.
    cur_dir = os.path.dirname(os.path.abspath(__file__))
    j2_env = Environment(loader=FileSystemLoader(cur_dir), trim_blocks=True)
    # prepare ranked list filenames: [query, db_list].
    db_ranked_fns = []
    db_ranked_labels = []
    for cur_ranked_ids in self.ranked_lists:
      db_ranked_fns.append(
          [self.db_items[x]["img_fn"] for x in cur_ranked_ids[:10]])
      db_ranked_labels.append(
          [self.db_items[x]["label"] for x in cur_ranked_ids[:10]])
    # render webpage.
    query_files = [x["img_fn"] for x in self.query_items]
    query_labels = [x["label"] for x in self.query_items]
    res_str = j2_env.get_template("match_vis_temp.html").render(
        test_name=res_name,
        query_fns=query_files[:100],
        query_labels=query_labels[:100],
        db_fns=db_ranked_fns[:100],
        db_labels=db_ranked_labels[:100],
        rank_accu=zip(accu_rank, top_k_accu))
    res_fn = os.path.join(self.res_dir, "match_vis.html")
    with open(res_fn, "w") as fn:
      fn.write(res_str)
      self.logger.info("results saved to {}".format(res_fn))

  def run_dev(self,
              img_folder,
              img_exts,
              load_items_from_db=False,
              compute_feats=False,
              compress_feats=False,
              train_compressor=False):
    """[DEV] All in one method to run pipeline.

    Be able to use the same data split with different
    features for comparison.
    """
    if load_items_from_db:
      self.load_items_from_db()
    else:
      self.read_data_from_files(img_folder, img_exts)
      self.load_items_from_db()
    if compute_feats:
      self.compute_feats()
    self.train_feat_compressor = train_compressor
    if compress_feats:
      self.compress_feats()
    self.run_match()
    self.evaluate()

  def start_dev(self,
                feature,
                feat_comp_batch_size=20,
                use_feat_compressor=True,
                compress_feat_dim=128,
                feat_compress_batch_size=200,
                match_protocol=search_common.MatchProtocol.EXACT,
                num_match_trees=100):
    """Set up dev environment.
    """
    # set up feature extractor.
    self.feat_extractor = feature
    self.target_feat_name = feature.feat_name
    self.target_feat_dim = feature.get_feat_dim()
    self.mode = PipelineMode.DEVELOPMENT
    self.feat_compute_batch_size = feat_comp_batch_size
    self.logger.info("starting feature extractor...")
    startt = time.time()
    self.feat_extractor.start()
    self.logger.info("feature extractor started, time: {}s.".format(time.time()
                                                                    - startt))
    # set up feature compressor.
    self.use_feat_compressor = use_feat_compressor
    self.feat_compress_batch_size = feat_compress_batch_size
    self.compress_feat_dim = compress_feat_dim
    if self.use_feat_compressor:
      self.feat_compressor = dim_reductor.IncrementalPCADimReducer(
          self.res_dir,
          self.compress_feat_dim,
          batch_size=self.feat_compress_batch_size)
      self.target_feat_dim = self.compress_feat_dim
    # set up matcher.
    self.match_protocol = match_protocol
    self.num_trees = num_match_trees
    if self.match_protocol == search_common.MatchProtocol.APPROXIMATE:
      self.matcher = simple_matchers.ANNMatcher(
          "{}_matcher".format(self.task_name), search_common.MatcherType.L2,
          self.target_feat_dim, self.num_trees, self.res_dir)
    else:
      self.matcher = simple_matchers.ExactMatcher(
          "{}_matcher".format(self.task_name), search_common.MatcherType.L2,
          self.res_dir)

  def start_deploy(self,
                   feature,
                   use_feat_compressor=True,
                   compress_feat_dim=128,
                   match_protocol=search_common.MatchProtocol.EXACT,
                   num_match_trees=100):
    """Set up and get ready for production.
    """
    # load featur extractor.
    self.feat_extractor = feature
    self.target_feat_name = feature.feat_name
    self.target_feat_dim = feature.get_feat_dim()
    self.mode = PipelineMode.DEPLOYMENT
    self.logger.info("starting feature extractor...")
    startt = time.time()
    self.feat_extractor.start()
    self.logger.info("feature extractor started, time: {}s.".format(time.time()
                                                                    - startt))
    # load feature compressor.
    self.use_feat_compressor = use_feat_compressor
    self.compress_feat_dim = compress_feat_dim
    if self.use_feat_compressor:
      self.logger.info("loading feature compressor...")
      self.feat_compressor = dim_reductor.IncrementalPCADimReducer(
          self.res_dir, self.compress_feat_dim)
      self.feat_compressor.save_load_model(to_save=False)
      self.logger.info("feature compressor loaded.")
      self.target_feat_dim = self.compress_feat_dim
    # load matcher.
    self.logger.info("loading matcher...")
    self.match_protocol = match_protocol
    self.num_trees = num_match_trees
    if self.match_protocol == search_common.MatchProtocol.APPROXIMATE:
      self.matcher = simple_matchers.ANNMatcher(
          "{}_matcher".format(self.task_name), search_common.MatcherType.L2,
          self.target_feat_dim, self.num_trees, self.res_dir)
      self.matcher.load_index()
    else:
      self.matcher = simple_matchers.ExactMatcher(
          "{}_matcher".format(self.task_name), search_common.MatcherType.L2,
          self.res_dir)
      self.load_feats_from_db(not self.use_feat_compressor)
    self.logger.info("matcher loaded.")

  def feat_post_process(self, feat):
    """Post processing feature.

    Override for custom process.

    Args:
      feat: 1D 32b float feature vector.
    Returns:
      post processing feature.
    """
    # reduce feature dimension.
    new_feat = feat
    if self.use_feat_compressor:
      if len(feat.shape) == 1:
        feat = feat.reshape((1, -1))
      new_feat = self.feat_compressor.transform(feat)
      new_feat = normalize(new_feat)
    return new_feat

  def search(self, img_rgb, res_num):
    """deploy purposed search.

    Args:
      img_rgb: input image query.
      res_num: how many results to retrieve.
    Returns:
      result metadata.
    """
    # compute feature.
    feat = self.feat_extractor.compute(img_rgb)
    # pose processing.
    new_feat = self.feat_post_process(feat)
    # search.
    query_items = {"metadata": [], "features": new_feat}
    ranked_list = self.matcher.match(query_items, top_nn=res_num)
    res_data = [self.matcher.db_metadata[x] for x in ranked_list[0]]
    return res_data



"""Tests for vision api.
"""

import unittest

import owl.data.img_tools as img_tools
import owl.third.aws as owl_aws
import owl.third.vision as vision_api


class VisionAPITester(unittest.TestCase):
  test_img_fn = "/mnt/Lab/imgs/1_25_25389.jpg"
  test_img_url = "https://static1.squarespace.com/static/58b4437f03596e617b3ebd07/58b5c7b8d2b857059b9b705b/58b5c7bdd1758ecc1d4a86ff/1488308168310/eg3.png"

  def test_clarifai(self):
    api_key = "c87f6abeb46f41f1baabb014626a4d23"
    api = vision_api.clarifai_api.ClarifaiAPI(api_key)
    res = api.label_img_general(img_fn=self.test_img_fn)
    print res
    # res = api.label_img_general(self.test_img_url)

  def test_rekognition(self):
    credentials = owl_aws.common.read_config_fn("../aws_config.json")
    api = vision_api.rekognition_api.RekognitionAPI(credentials["access_key"],
                                                    credentials["secret_key"],
                                                    credentials["region"])
    img_bin = img_tools.read_img_bin(self.test_img_fn)
    res = api.label_img(img_bin)
    print res

  def test_google(self):
    pass


if __name__ == "__main__":
  unittest.main()



"""Clarifai api manager.

Need to create account and obtain api key.
"""

import json

from clarifai import rest
from clarifai.rest import ClarifaiApp


class ClarifaiAPI(object):
  """Class to manage clarifai api.
  """
  client = None

  def __init__(self, api_key):
    self.client = ClarifaiApp(api_key=api_key)

  def label_img_general(self,
                        img_url=None,
                        img_base64=None,
                        img_bin=None,
                        img_fn=None,
                        top_k=10):
    """Predict image labels using general model.
    
    Use any non-empty value as image.
    """
    model = self.client.models.get("general-v1.3")
    if img_url is not None:
      res = model.predict_by_url(img_url, max_concepts=top_k)
    if img_base64 is not None:
      res = model.predict_by_base64(img_base64, max_concepts=top_k)
    if img_bin is not None:
      res = model.predict_by_bytes(img_bin, max_concepts=top_k)
    if img_fn is not None:
      res = model.predict_by_filename(img_fn, max_concepts=top_k)
    new_res = []
    for item in res["outputs"][0]["data"]["concepts"]:
      new_res.append((item["name"], item["value"]))
    return new_res



import json

from owl.processors.recognition import aws_face


class AWSFaceTester(object):
  aws_configs = None

  def __init__(self):
    with open("./aws_config.json", "r") as f:
      self.aws_configs = json.load(f)

  def test_face_detect(self):
    face_det = aws_face.AWSFaceDetector(self.aws_configs["access_key"],
                                        self.aws_configs["secret_key"],
                                        self.aws_configs["region"])
    test_img_path = ""


"""Basic recognition processor objects.

All the object structures are agnostic to actual algorithm.
These are prediction interfaces.
"""

import abc
import random

from owl.data.img_obj import OwlImage
""" Image classification """


class ImgClfPred(object):
  """Class for image classification result.
  """
  img_ref = None
  scored_labels = None

  def __init__(self, label_scores=None):
    """Create object from a list of scored labels.

    Args:
      scored_labels: [(label_name, label_score)].
      Compatible with deepmodels classifier output.
    """
    if label_scores:
      self.scored_labels = []
      for label_score in label_scores:
        self.scored_labels.append(
            (label_score[0], format(label_score[1], ".4f")))

  def to_json(self):
    return {"img_ref": self.img_ref, "labels": self.scored_labels}


class ImgClassifierBase(object):
  """Base class for image classification processors.
  """
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def __init__(self):
    pass

  @abc.abstractmethod
  def prepare(self):
    """Prepare prediction model.
    """
    pass

  @abc.abstractmethod
  def predict(self, img):
    """Take input of OwlImg and output ImgClfPred.

    Returns: ImgClfRes object.
    """
    assert isinstance(img, OwlImage)


""" Image detection """


class ImgDetPred(object):
  """Class for image detection result.
  """
  img_ref = None
  # list of bounding box: <xmin, ymin, width, height>.
  bboxes = []
  box_labels = []
  box_scores = []


class ImgDetectorBase(object):
  """Base class for image detection processors.
  """
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def __init__(self):
    pass

  @abc.abstractmethod
  def prepare(self):
    """Prepare prediction model.
    """
    pass

  @abc.abstractmethod
  def predict(self, img):
    pass

  def vis_preds(self, img, pred):
    """Draw detection boxes on image.
    """
    line_colors = []
    texts = []
    for idx in range(len(pred.bboxes)):
      cur_color = (random.randint(0, 255), random.randint(0, 255),
                   random.randint(0, 255))
      line_colors.append(cur_color)
      texts.append(str(idx))
    new_img = img.draw_boxes(
        pred.bboxes, line_colors=line_colors, line_width=3, texts=texts)
    return new_img



"""OOP interface for S3.
"""

from cStringIO import StringIO

from owl.third.aws import common as aws_common
from owl.data.img_obj import OwlImage


class S3IO(aws_common.AWSService):
  """Class for managing s3 tasks.
  """
  bucket = None
  bucket_name = None

  def __init__(self, access_key, secret_key, region):
    """Create s3 client.
    """
    self.init_aws("s3", access_key, secret_key, region)

  def use_bucket(self, bucket_name):
    """Use a specified bucket.

    If the bucket doesn't exist, create one.
    """
    self.bucket = self.resource.create_bucket(Bucket=bucket_name)
    self.bucket_name = bucket_name

  def upload_img_bin(self,
                     img_bin,
                     img_format="png",
                     key_prefix="",
                     make_public=False):
    """Upload image binary data to bucket.

    Args:
      img_bin: image binary data.
      img_format: type of image, only support jpy or png.
      key_prefix: prefix of image key, used to create folders.
      make_public: whether to make it publically accessible.
    Returns:
      image object key.
    """
    img_format = img_format.lower()
    assert img_format in ["png", "jpg", "jpeg"
                          ], "img_format must be either png, jpg, jpeg."
    assert self.bucket_name
    # make sure no duplicate images are saved.
    img_obj = OwlImage(img_bin=img_bin)
    key_prefix = key_prefix.strip("/")
    img_key_name = "{}/{}.{}".format(key_prefix,
                                     img_obj.get_base64_sha_encoding(),
                                     img_format)
    img_key_name = img_key_name.strip("/")
    upload_args = {}
    if img_format in ["jpg", "jpeg"]:
      upload_args["ContentType"] = "image/jpeg"
    if img_format == "png":
      upload_args["ContentType"] = "image/png"
    if make_public:
      upload_args["ACL"] = "public-read"
    self.bucket.upload_fileobj(
        StringIO(img_bin), img_key_name, ExtraArgs=upload_args)
    return img_key_name

  def upload_img_base64(self, img_base64, key_prefix=""):
    """Save image data to bucket.

    sha1 is used as key for image data.

    Args:
      img_base64: base64 string of image.
    Returns:
      image sha hash as key.
    """
    img_obj = OwlImage(img_base64=img_base64)
    img_sha = img_obj.get_base64_sha_encoding()
    self.bucket.put_object(
        Key="{}/{}".format(key_prefix, img_sha), Body=img_base64)
    return img_sha

  def generate_obj_url(self, obj_key):
    """Generate url of object.

    Args:
      obj_key: key to object in the bucket.

    Returns:
      the url for object.
    """
    img_url = "{}/{}/{}".format(self.client.meta.endpoint_url,
                                self.bucket_name, obj_key)
    return img_url

  def generate_presigned_url(self, obj_key):
    """Generate a temporary url.

    Args:
      obj_key: key of object in the current bucket.

    Returns:
      presigned url for the target object.
    """
    url = self.client.generate_presigned_url(
        ClientMethod='get_object',
        Params={'Bucket': self.bucket_name,
                'Key': obj_key})
    return url

  def get_object(self, obj_key):
    """Get object binary data from bucket.

    Args:
      obj_key: object key.
    Returns:
      object data.
    """
    res = self.client.get_object(Bucket=self.bucket_name, Key=obj_key)
    obj_data = res["Body"].read()
    return obj_data



"""Test mongodb manager.
"""

import time

import unittest
from owl.data import mongo_manager

class MongoDBTester(unittest.TestCase):
  db = mongo_manager.MongoManager()

  def test_query_speed(self):
    self.db.connect(db_name="eyestyle_production", collection_name="products_2016_02")
    startt = time.time()
    items = self.db.query(attribute="category", value_list=["casual-pants"], limit=0)
    print "query category time: {}s, items: {}".format(time.time()-startt, len(items))

if __name__ == "__main__":
  unittest.main()



"""Google vision api manager.
"""


class GoogleVisionAPI(object):
  """Class to manage google cloud vision api.
  """
  client = None

  def __init__(self):
    pass

  def label_img(self, img_base64):
    """Label image with general lables.
    """
    pass



"""Common data definition.
"""

import abc

import numpy as np


class FeatType(object):
  """Feature type in python.
  """
  # CNN feature.
  CNN_VGGS = 0
  CNN_VGGS_SHORT = 1
  CNN_VGG16 = 2
  CNN_VGG16_SHORT = 3
  CNN_ALEX = 4
  CNN = 6
  # color feature.
  COLOR_CLUSTER = 5
  COLOR_HIST_LINEAR = 6
  COLOR_HIST_JOINT = 7
  # extra.
  CUSTOM = 100


class DistType(object):
  """Distance type.
  """
  L2 = 0
  ANGULAR = 1
  CUSTOM = 2


class ColorSpace(object):
  """Color space name.
  """
  RGB = 0
  BGR = 1
  HSV = 2
  LAB = 3


class FeatParamsBase(object):
  """Base class for feature params.
  """
  feat_type = 0
  feat_name = ""


# map python feature type to c++ feature name.
FeatTypeToNameMapper = {
    FeatType.CNN_VGGS: "cnn_vggs",
    FeatType.CNN_VGGS_SHORT: "cnn_vggs_short",
    FeatType.CNN_VGG16: "cnn_vgg16",
    FeatType.CNN_VGG16_SHORT: "cnn_vgg16_short",
    FeatType.CNN_ALEX: "cnn_alex"
}


def map_name_to_feat_type(feat_name):
  """Map a feature name string to feature type.

  Args:
    feat_name: feature name string.
  """
  for py_type, name in FeatTypeToNameMapper.iteritems():
    if feat_name == name:
      return py_type
  raise ValueError("feat name not exist")


def is_cnn_feat(feat_name):
  """Check if a feature name is cnn feature.

  Args:
    feat_name: feature name string.
  """
  if feat_name in FeatTypeToNameMapper.values():
    return True
  else:
    return False


class BaseFeatParams(object):
  """Base class for feature extractor params.

  Put feature related params here.
  """
  pass


class BaseFeatExtractor(object):
  """Base class for feature extraction.

  Generic usage of a feature extractor:
  1) create: my_extractor = MyFeature(feat_params)
  2) start (load trainable feature, e.g. dnn, bow): my_extractor.start()
  3) set input: my_extractor.set_input(color_img) or
  4) compute feature: my_extractor.compute(color_img, mask)
  5) compare feature: my_extractor.calc_dist(feat1, feat2)

  Attributes:
    feat_tpye: type of feature.
    feat_name: name of feature.
    feat_params: feature parameters.
    raw_img_bgr: original bgr image.
    internal_img: intermediate image representation, e.g. gray image.
  """
  __metaclass__ = abc.ABCMeta

  feat_type = None
  feat_name = None
  feat_dim = 0
  feat_params_ = None
  raw_img_rgb = None
  internal_img = None

  @abc.abstractmethod
  def __init__(self, feat_params=None):
    """Initialize feature extractor.

    Override to set property of the extractor.

    Args:
      feat_params: parameters for feature extractor.
    """
    pass

  def set_input(self, rgb_img):
    """Set input image and preprocess it.

    This is useful for repeated feature extraction from
    the same image with shared middle data.

    Args:
      rgb_img: 3D numpy array as rgb image.
    """
    self.raw_img_rgb = rgb_img

  def start(self):
    """Prepare feature extractor.
    """
    print "feature {} started".format(self.feat_name)

  @abc.abstractmethod
  def compute(self, rgb_img=None, mask=None):
    """Compute image feature.

    Args:
      rgb_img: input image. If none, use internal image.
      mask: pixel mask. non-zero indicates valid.
    Returns:
      1D ndarray feature vector.
    """
    pass

  def compute_batch(self, rgb_imgs, masks=None):
    """Compute image features in batch.

    Args:
      rgb_imgs: a list of rgb_img.
    Returns:
      batch features as 2D matrix, each row is a feature vector.
    """
    feats = []
    for idx, rgb_img in enumerate(rgb_imgs):
      if masks is not None:
        cur_feat = self.compute(rgb_img, masks[idx])
      else:
        cur_feat = self.compute(rgb_img)
      feats.append(cur_feat)
    return np.stack(feats)

  @abc.abstractmethod
  def calc_dist(self, feat1, feat2):
    """Compute distance between two feature vectors.

    Used for custom distance metric.

    Args:
      feat1: first feature vector.
      feat2: second feature vector.
    Returns:
      float distance value.
    """
    pass

  def get_feat_dim(self):
    """Return feature dimenion.
    """
    return self.feat_dim



import os
import random

from owl.data.img_obj import OwlImage
from owl.processors.recognition.common import ImgDetPred, ImgDetectorBase
from owl.pipelines.recognition.img_det.pipeline import ImgDetectionPipeline


class DummyImgDet(ImgDetectorBase):
  def __init__(self):
    pass

  def prepare(self):
    pass

  def predict(self, imgs):
    pass


class TestImgDetPipeline(object):
  img_dir = "/mnt/Lab/imgs/salientobj/data/"

  def test_display(self):
    dummy_preds = []
    imgs = os.listdir(self.img_dir)
    imgs = [os.path.join(self.img_dir, x) for x in imgs]
    for i in range(30):
      cur_pred = ImgDetPred()
      cur_pred.box_labels = ["dog", "cat", "cow", "apple", "pear", "peach"]
      cur_pred.box_scores = [random.random()] * 6
      cur_pred.bboxes = [(10, 10, 120, 110)] * 6
      cur_img = OwlImage(fp=imgs[i])
      cur_img.img_arr = cur_img.draw_box(cur_pred.bboxes[0], line_width=3)
      cur_pred.img_ref = cur_img.to_datauri()
      dummy_preds.append(cur_pred)
    pp = ImgDetectionPipeline(DummyImgDet())
    pp.show(dummy_preds, "/mnt/Lab/results/owlpipelines/imgdet.html")



"""Service to detect objects in image.
"""

import os
from tqdm import tqdm
from jinja2 import Environment, FileSystemLoader

from owl.processors.recognition.common import ImgDetPred, ImgDetectorBase


class ImgDetectionPipeline(object):
  """Class for image classification pipeline.
  """
  model = None

  def __init__(self, model_):
    """Set up prediction model.
    """
    assert isinstance(model_, ImgDetectorBase), "incorrect model type."
    self.model = model_

  def prepare(self):
    self.model.prepare()

  def predict(self, input_imgs):
    """Make prediction for images.

    Args:
      input_imgs: a list of owlimage objects.
    """
    print "start predicting images..."
    preds = []
    for cur_img in tqdm(input_imgs):
      cur_pred = self.model.predict(cur_img)
      vis_img = self.model.vis_preds(cur_img, cur_pred)
      vis_img.show("detect")
      cur_pred.img_ref = vis_img.to_datauri()
      preds.append(cur_pred)
    return preds

  def show(self, preds, save_fn):
    """Show predictions in a web page.
    """
    assert save_fn[-4:] == "html", "save file must be an html file."
    cur_dir = os.path.dirname(os.path.abspath(__file__))
    j2_env = Environment(loader=FileSystemLoader(cur_dir), trim_blocks=True)
    res_str = j2_env.get_template("./app/index.html").render(
        name="test", preds=preds)
    with open(save_fn, "w") as fn:
      fn.write(res_str)
      print("results saved to {}".format(save_fn))


if __name__ == "__main__":
  pass


import unittest

import objectpath

from owl.data import common


class CommonToolsTester(unittest.TestCase):
  def test_split_data(self):
    img_folder = "/mnt/DataHouse/Fashion/EyeStyle/cnn_test/pipeline/test1/data/"
    img_fns, img_labels, label_names = common.list_img_files(img_folder)
    self.assertEqual(len(img_fns), len(img_labels))
    db_ids, _, query_ids = common.split_train_val_test(
        img_labels, train_ratio=0.8, test_ratio=0.2)
    self.assertAlmostEqual(len(db_ids), 0.8 * len(img_labels))

  def test_jsonpath_val(self):
    dummy_json_obj = {
      "brand": {
        "name": [{
          "3": "test"
        }, {
          "key": "ds"
        }]
      }
    }
    obj_tree = objectpath.Tree(dummy_json_obj)
    vals = obj_tree.execute("$.brand.name[0].'3'")
    print vals


if __name__ == "__main__":
  unittest.main()


"""Tests for visualsearch pipeline.
"""

import unittest

from owl.pipelines.visualsearch import visualsearch


class VisualSearchTester(unittest.TestCase):
  res_dir = "/mnt/DataHouse/Fashion/EyeStyle/cnn_test/pipeline/test1/res_test/"
  img_dir = "/mnt/DataHouse/Fashion/EyeStyle/cnn_test/pipeline/test1/data/"
  pipe = visualsearch.OneFeatVisualSearcher("vs_test", res_dir)

  def test_file_loading(self):
    self.pipe.read_data_from_files(self.img_dir, ["*.jpg"])

  def test_db_loading(self):
    self.pipe.load_items_from_db()

  def test_dev(self):
    self.pipe.start_dev(None)
    self.pipe.run_dev(
        self.img_dir, ["*.jpg"],
        load_items_from_db=False,
        compute_feats=True,
        compress_feats=True,
        train_compressor=True)

  def test_deploy(self):
    self.pipe.start_deploy(None)
    query_img = None
    res_data = self.pipe.search(query_img, 1000)


if __name__ == "__main__":
  unittest.main()



"""Database management for web api.

NOTE: for linux dev, use mongobooster to inspect database content visually.
"""

import pymongo
from pymongo import MongoClient
from bson.objectid import ObjectId


def objectid_to_str(object_id):
  return str(object_id)


def str_to_objectid(id_str):
  return ObjectId(id_str)


class MongoManager(object):
  """MongoDB manager.
  """

  # data
  client = None
  db = None
  collection = None

  def __init__(self):
    pass

  def __del__(self):
    self.disconnect()

  def is_connection_valid(self):
    """Check if connection is valid.
    """
    if self.client == None or \
    self.db == None or \
    self.collection == None:
      print "db connection invalid"
      return False
    else:
      return True

  def connect(self,
              user=None,
              pwd=None,
              host=None,
              port=None,
              db_name=None,
              collection_name=None):
    """Connect to mongodb server.

    Args:
      user: user to login.
      pwd: pwd to login.
      host: server host.
      port: port to use.
      db_name: database to use.
      collection_name: collection to use.
    """
    self.disconnect()
    # make url.
    db_url = "mongodb://"
    if user is not None:
      db_url += user
    if pwd is not None:
      db_url += ":{}@".format(pwd)
    db_url += host if host is not None else "127.0.0.1"
    db_url += ":{}".format(port) if port is not None else ":27017"
    # get client.
    print "connecting to mongodb: {}".format(db_url)
    self.client = MongoClient(db_url)
    if db_name is not None:
      self.db = self.client[db_name]
      print "use db: {}".format(db_name)
    else:
      self.db = self.client["test"]
    if collection_name is not None:
      self.collection = self.db[collection_name]
      print "use collection: {}".format(collection_name)
    else:
      self.collection = self.db["test"]
    print "mongodb connected"

  def delete_db(self, db_name):
    """Delete database.
    """
    self.client.drop_database(db_name)
    print "removed db: {}".format(db_name)

  def delete_collection(self):
    """delete collection for current db.
    """
    self.collection.drop()
    print "current collection dropped."

  def switch_db_collection(self, db_name, collection_name):
    """Select a specific collection.
    """
    assert self.client is not None, "db is not connected."
    self.db = self.client[db_name]
    self.collection = self.db[collection_name]
    print "db switch to: {}.{}".format(db_name, collection_name)

  def disconnect(self):
    """disconnect from server.
    """
    if self.client is not None:
      self.client.close()

  def create_index(self, index_attribute):
    """Create index on current collection.

    Args:
      index_attribute: attribute name to index.
    """
    self.collection.create_index(index_attribute)
    print "index {} created.".format(index_attribute)

  def add(self, obj):
    """Add a new document.

    Args:
      obj: document to insert.
    Returns:
      object_id of the inserted document.
    """
    self.collection.insert_one(obj)
    return obj["_id"]

  def add_many(self, obj_list):
    """Add a list of objects together.

    Args:
      obj_list: documents to insert.
    """
    self.collection.insert_many(obj_list)

  def get_scan_cursor(self):
    """Return cursor to all documents.
    """
    cursor = self.collection.find()
    return cursor

  def query(self, attribute="_id", value_list=None, limit=0):
    """query a set of objects using certain attribute.

    Args:
      attribute: name of the attribute.
      value_list: list of values to search for the attribute.
    Returns:
      found objects.
    """
    if not self.is_connection_valid():
      raise pymongo.errors.CollectionInvalid()
    res = []
    for val in value_list:
      docs = self.collection.find({attribute: val}, limit=limit)
      res.extend(docs)
    return res

  def update(self, obj_id, new_key_vals):
    """Update an object.

    Args:
      obj_id: object id object.
      new_key_vals: dict of new key value pairs.
    """
    try:
      _ = self.collection.update_one({"_id": obj_id}, {"$set": new_key_vals})
    except pymongo.errors.PyMongoError as ex:
      print "error {}".format(ex)


if __name__ == '__main__':
  dm = MongoManager()
  config_fn = 'E:/Projects/Github/VisualSearchEngine/Owl/Settings/engine_settings2.json'



"""Web api classes.
"""

from flask import request, jsonify

from owl.data.img_obj import OwlImage
from owl.services import web_api
from owl.processors.recognition import common


class ImgClassifierWebAPI(web_api.WebAPIBase):
  """Generic web api for image classification.
  """
  classifier = None

  def __init__(self, api_name, clf_obj):
    """Initialize api.
    """
    super(ImgClassifierWebAPI, self).__init__(api_name)
    assert isinstance(clf_obj, common.ImgClassifierBase)
    self.classifier = clf_obj
    # add resource.
    self.add_resource("/predict", ["GET", "POST"], self.predict)

  def prepare(self):
    self.classifier.prepare()
    self.started = True

  def parse_clf_request(self, req_json):
    assert "img_base64" in req_json
    return req_json["img_base64"]

  def predict(self):
    """Perform prediction on request image.
    """
    try:
      if request.method == "GET":
        if self.started:
          return "classification service is on"
        else:
          return "classification service is not on"
      if request.method == "POST":
        req_data = request.get_json()
        img_base64 = self.parse_clf_request(req_data)
        img_obj = OwlImage(img_base64=img_base64)
        clf_pred = self.classifier.predict(img_obj)
        res_obj = web_api.make_success_res_obj(clf_pred.to_json())
        print res_obj
        return jsonify(res_obj), 200
    except Exception as ex:
      print "[{}] {}".format(self.__class__.__name__, ex)
      return jsonify(web_api.make_error_res_obj(ex.message)), 500

  def update_predict_path(self, new_path):
    """Change default predict url path.
    """
    # TODO(jiefeng): figure out how to remove the existing one.
    self.add_resource(new_path, ["GET", "POST"], self.predict)


class ImgDetectorWebAPI(web_api.WebAPIBase):
  """Generic web api for image detection.
  """
  detector = None

  def __init__(self, api_name, detector_obj):
    """Initialize api.
    """
    super(ImgDetectorWebAPI, self).__init__(api_name)
    assert isinstance(clf_obj, common.ImgDetectorBase)
    self.detector = detector_obj
    # add resource.
    self.add_resource("/predict", ["GET", "POST"], self.predict)

  def prepare(self):
    self.detector.prepare()
    self.started = True

  def predict(self):
    """Perform prediction on request image.
    """
    try:
      if request.method == "GET":
        if self.started:
          return "detection service is on"
        else:
          return "detection service is not on"
      if request.method == "POST":
        req_data = request.get_json()
        pass

    except Exception as ex:
      return jsonify(web_api.make_error_res_obj(ex.message)), 500

  def update_predict_path(self, new_path):
    """Change default predict url path.
    """
    self.add_resource(new_path, ["GET", "POST"], self.predict)


"""AWS ReKognition api manager.

Need aws account and assign rekognition permission.
"""

import json

from owl.data import img_tools
from owl.third.aws import common as aws_common


class AWSRekognitionAPI(aws_common.AWSService):
  """Class for managing aws rekognition api.
    """

  def __init__(self, access_key, secret_key, region):
    """Create rekognition client.
    """
    self.init_aws("rekognition", access_key, secret_key, region)

  def label_img(self, img_bin, top_k=10):
    """Predict image labels.

    Args:
      img_bin: binary image data.
      top_k: top k predictions.
    Returns:
      a list of predicted labels and probability, e.g. [("dog", 0.65)]
    """
    res = self.client.detect_labels(Image={"Bytes": img_bin}, MaxLabels=top_k)
    new_res = []
    for item in res["Labels"]:
      new_res.append((item["Name"], item["Confidence"]))
    return new_res

  def detect_faces(self, img_bin):
    """Detect faces in given image.

    Api ref: http://boto3.readthedocs.io/en/latest/reference/services/rekognition.html#Rekognition.Client.detect_faces

    Args:
      img_bin: image binary data.
    
    Returns:
      a list of face details.
    """
    res = self.client.detect_faces(
        Image={"Bytes": img_bin}, Attributes=["ALL"])
    res_data = res["FaceDetails"]
    return res_data

  def detect_celebrities(self, img_bin):
    """Detect celebrities in a given image.

    Args:
      img_bin: image binary data.

    Returns:
      a list of detected celebrities details.
    """
    res = self.client.recognize_celebrities(Image={"Bytes": img_bin})
    return res["CelebrityFaces"]

  def detect_text(self, img_bin):
    """Detect text in image.

    Args:
      img_bin: binary image data.

    Returns:
      a list of detected text.
    """
    res = self.client.detect_text(Image={"Bytes": img_bin})
    text_res = []
    for det_res in res["TextDetections"]:
      cur_res = {}
      cur_res["text"] = det_res["DetectedText"]
      cur_res["bbox"] = det_res["Geometry"]["BoundingBox"]
      cur_res["score"] = det_res["Confidence"]
      text_res.append(cur_res)
    return text_res


if __name__ == "__main__":
  with open("../aws_config.json", "r") as f:
    config = json.load(f)
  api = AWSRekognitionAPI(config["access_key"], config["secret_key"],
                          config["region"])
  img = img_tools.read_img_bin("../../assets/bird.jpg")
  labels = api.label_img(img)
  texts = api.detect_text(img)
  print labels
  print texts


import os
from setuptools import setup, find_packages

try:
  import GPUtil
except:
  # install package.
  os.system("pip install gputil")
  import GPUtil


def if_gpu_exist():
  deviceIDs = GPUtil.getAvailable(
      order="first", limit=1, maxLoad=0.1, maxMemory=0.1)
  return len(deviceIDs) > 0


with open("./configs/requirements.txt", "r") as f:
  dep_packages = f.readlines()
  # remove local install.
  dep_packages = [x.strip() for x in dep_packages if not x.startswith("-e")]
  for dep_id, dep in enumerate(dep_packages):
    # select version of tensorflow.
    if dep in ["tensorflow-gpu", "tensorflow"]:
      package_name, package_ver = dep.split("==")
      if if_gpu_exist():
        package_name = "tensorflow-gpu"
      else:
        package_name = "tensorflow"
      dep_packages[dep_id] = "{}=={}".format(package_name, package_ver)

setup(
    name="projectowl",
    version="0.3.9",
    description="python app library",
    url="https://flyfj@bitbucket.org/flyfj/owl.git",
    author="Jie Feng",
    author_email="jiefengdev@gmail.com",
    packages=find_packages("./"),
    install_requires=dep_packages,
    include_package_data=True,
    zip_safe=False)



"""Class to generate different types of color descriptors.
"""

import numpy as np
from sklearn.cluster import KMeans

import cv2

from owl.data import img_tools
from owl.features import common


class ColorParams(common.BaseFeatParams):
  """Color specific parameters.
  """
  pass


class ColorClusters(common.BaseFeatExtractor):
  """Color cluster descriptors.
  """
  # each element is a tuple of pixel value.
  color_palette = []

  def __init__(self, feat_params):
    super(ColorClusters, self).__init__(feat_params)
    self.feat_name = "color_cluster"
    self.feat_type = common.FeatType.COLOR_CLUSTER
    self.feat_params_ = feat_params

  def set_input(self, rgb_img):
    self.internal_img = np.float32(rgb_img)
    # convert to LAB.
    self.internal_img /= 255

  def compute(self, rgb_img=None, mask=None):
    # set input.
    if rgb_img is not None:
      self.set_input(rgb_img)
    # reshape img pixels.
    if mask is None:
      pixels = self.internal_img.reshape((-1, 3))
    else:
      imgw, imgh, _ = self.internal_img.shape
      # filter valid pixels.
      pixels = [
          self.internal_img[row, col]
          for row, col in zip(range(imgh), range(imgw)) if mask[row, col] > 0
      ]
    # do clustering.
    num_cluster = 8
    self.feat_dim = num_cluster * 4
    # sklearn kmeans
    #clr = KMeans(n_clusters=num_cluster)
    #clr.fit(pixels)
    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 10, 1.0)
    # fit.
    _, labels, centers = cv2.kmeans(pixels, num_cluster, None, criteria, 10,
                                    cv2.KMEANS_PP_CENTERS)
    # compute weight.
    weights = np.zeros((num_cluster, 1), np.float32)
    for cls_id in labels:
      weights[cls_id] += 1
    weights /= len(pixels)
    # rank cluster centers based on weights from high to low.
    # order: (pixel, weight).
    weighted_cls = np.append(centers, weights, axis=1)
    sorted_cls = sorted(weighted_cls, key=lambda cls: cls[3], reverse=True)
    # form feature vector by concatenating weights and cluster centers.
    feat = []
    self.color_palette = []
    for cluster in sorted_cls:
      feat.extend([cluster[3], cluster[0], cluster[1], cluster[2]])
      self.color_palette.append((int(cluster[0] * 255), int(cluster[1] * 255),
                                 int(cluster[2] * 255)))
    feat = np.array(feat).reshape((1, -1))
    return feat

  def vis_clusters(self):
    """Visualize clusters and save to image.
    """
    num_cluster = len(self.color_palette)
    canvas = np.zeros((50, 51 * num_cluster, 3), np.uint8)
    for idx, cls_center in enumerate(self.color_palette):
      pt1 = (idx * 50, 0)
      pt2 = (idx * 50 + 50, 50)
      cv2.rectangle(canvas, pt1, pt2, cls_center, -1)
    img_tools.write_img_arr(canvas, "palette.png")

  def calc_dist(self, feat1, feat2):
    clusters1 = []
    clusters2 = []
    assert len(feat1) == len(
        feat2), "two features must have the same dimension"
    cls_num = len(feat1) / 4
    for i in range(cls_num):
      # first element is weight.
      cls1 = feat1[i * 4:i * 4 + 4]
      cls2 = feat2[i * 4:i * 4 + 4]
      clusters1.append(cls1)
      clusters2.append(cls2)
    # compute pairwise cluster distance.
    cls_dists = np.zeros((cls_num, cls_num), np.float32)
    for id1, cls1 in enumerate(clusters1):
      for id2, cls2 in enumerate(clusters2):
        cur_dist = np.linalg.norm(cls1[1:4] - cls2[1:4], 2)
        cur_dist = min(1, cur_dist / MAX_DIST)
        cls_dists[id1, id2] = (id2, cur_dist)
      # sort distance from low to high.
      cls_dists[id1] = sorted(
          cls_dists[id1], key=lambda dist_unit: dist_unit[1])
    # temporary weight values during composition.
    tempw1 = [cls[0] for cls in clusters1]
    tempw2 = [cls[0] for cls in clusters2]
    # cost to compose 1 (query) using 2 (db).
    dist12 = 0
    for id1 in range(cls_num):
      for id2 in range(cls_num):
        best_id = cls_dists[id1, id2][0]
        if tempw2[best_id] <= 0:
          continue
        to_fill_amount = min(tempw1[id1], tempw2[best_id])
        dist12 += cls_dists[id1, id2][1] * to_fill_amount
        # update weight.
        tempw1[id1] -= to_fill_amount
        tempw2[best_id] -= to_fill_amount
        if tempw1[id1] <= 0:
          break

    return dist12



import os
import random

from owl.processors.recognition.common import ImgClfPred, ImgClassifierBase
from owl.pipelines.recognition.img_clf.pipeline import ImgClassificationPipeline


class DummyImgClf(ImgClassifierBase):
  def __init__(self):
    pass

  def prepare(self):
    pass

  def predict(self, imgs):
    pass


class TestImgClfPipeline(object):
  img_dir = "/mnt/Lab/imgs/salientobj/data/"

  def test_display(self):
    dummy_preds = []
    imgs = os.listdir(self.img_dir)
    imgs = [os.path.join(self.img_dir, x) for x in imgs]
    for i in range(30):
      cur_pred = ImgClfPred()
      cur_pred.img_ref = imgs[i]
      cur_pred.label_names = ["dog", "cat", "cow", "apple", "pear", "peach"]
      cur_pred.label_scores = [random.random()] * 6
      dummy_preds.append(cur_pred)
    pp = ImgClassificationPipeline(DummyImgClf())
    pp.show(dummy_preds, "/mnt/Lab/results/owlpipelines/imgclf.html")



"""Generic interface for feature extraction.
"""

import os
import cv2

import commentjson
from owl.features import common


class GenericObjFeatExtractor(object):
  """Python interface for feature extraction.

  Generic interface for extracting different features for an object.
  """

  # list of feature extractors.
  feat_extractors = {}
  raw_img_rgb = None
  inside_img = None
  config_fn = ""

  def __init__(self, config_fn="", features=None):
    """Create extractor.

    Supports init from config file or passed extractors.
    If both are present, combine them all.

    Args:
      config_fn: json file with feature configuration.
    Used for built-in features.
      features: a list of custom defined features.
    """
    if config_fn != "":
      self.init_with_fn(config_fn)
    if features is not None:
      self.init_from_features(features)

  def init_with_fn(self, config_fn):
    """Create extractor from configure file.

    Args:
      config_fn: json configuration file.
    """
    self.config_fn = config_fn

    # init python part
    with open(config_fn, 'r') as f:
      config = commentjson.load(f)

    # extract features
    feat_configs = config['engine']['features']
    for feat_config in feat_configs:
      if not feat_config['enabled']:
        continue

      if IsCNNFeat(feat_config['type']):
        # create params
        cur_params = CNNParams()
        cur_params.feat_name = feat_config['type'].encode("utf-8")
        print 'using feature: {}'.format(cur_params.feat_name)
        cur_params.feat_type = MapNameToFeatType(cur_params.feat_name)
        self.feat_params.append(cur_params)
        # create feature
        cnn_feat_computer = DeepCNNFeature()
        cnn_feat_computer.Init(cur_params)
        cnn_feat_computer.Start()
        self.feat_extractors.append(cnn_feat_computer)

    print "feature config loaded"

  def init_from_features(self, features):
    """Create extractor from custom features.

    Args:
      features: a list of custom defined feature extractors.
    """
    # add features.
    for feature in features:
      assert issubclass(
          type(feature), common.BaseFeatExtractor
      ), "feature extractor not a subclass of BaseFeatExtractor"
      self.feat_extractors[feature.feat_name] = feature

  def start(self):
    """Start all feature extractors.
    """
    for feature in self.feat_extractors.values():
      feature.start()

  def set_input(self, rgb_img):
    """Set image input.
    """
    for feature in self.feat_extractors.values():
      feature.set_input(rgb_img)

  def compute(self, rgb_img=None, mask=None):
    """Compute all features.

    Args:
      rgb_img: color input image.
      mask: image mask to select ROI.
    Returns:
      a dictionary of features.
    """
    feats = {}
    for feature in self.feat_extractors.values():
      cur_feat = feature.compute(rgb_img, mask)
      feats[feature.feat_name] = cur_feat
    return feats

  def compute_batch(self, rgb_imgs, masks=None):
    """Compute batch feature.
    """
    feats = {}
    for feature in self.feat_extractors.values():
      cur_feats = feature.compute_batch(rgb_imgs, masks)
      feats[feature.feat_name] = cur_feats
    return feats

  def calc_dist(self, feats1, feats2, weights=None):
    """Combined distance for feature sets.

    Args:
      feats1: feature set 1.
      feats2: feature set 2.
      weights: weight for each feature.
    Returns:
      final feature set distance.
    """
    total_dist = 0
    if weights is not None:
      assert len(weights) == len(
          feats1.keys), "inconsistent weight dimension with features"
    assert len(feats1.keys) == len(
        feats2.keys), "inconsistent feature set number"
    for feat_name in feats1.keys():
      total_dist += self.feat_extractors[feat_name].calc_dist(
          feats1[feat_name], feats2[feat_name])
    return total_dist


if __name__ == "__main__":
  pass



"""Face processors based on aws rekognition.
"""

import json

from owl.data.img_obj import OwlImage
from owl.processors.recognition import common
from owl.third.vision.rekognition_api import AWSRekognitionAPI
from owl.pipelines.recognition.img_det.pipeline import ImgDetectionPipeline


class AWSFaceDetector(common.ImgDetectorBase):
  """Face detector using aws.
  """
  engine = None

  def __init__(self, access_key, secret_key, region):
    self.engine = AWSRekognitionAPI(access_key, secret_key, region)

  def prepare(self):
    pass

  def predict(self, img):
    raw_res = self.engine.detect_faces(img.to_binary())
    new_res = common.ImgDetPred()
    new_res.img_ref = img.to_datauri()
    for face in raw_res:
      face_box = (int(face["BoundingBox"]["Left"] * img.width()),
                  int(face["BoundingBox"]["Top"] * img.height()),
                  int(face["BoundingBox"]["Width"] * img.width()),
                  int(face["BoundingBox"]["Height"] * img.height()))
      new_res.bboxes.append(face_box)
      new_res.box_labels.append(face["Gender"]["Value"])
      new_res.box_scores.append(face["Confidence"])
    return new_res


if __name__ == "__main__":
  with open("../../third/aws_config.json", "r") as f:
    config = json.load(f)
  face_detector = AWSFaceDetector(config["access_key"], config["secret_key"],
                                  config["region"])
  img = OwlImage(fp="../../assets/group_faces.jpg")
  det_pipeline = ImgDetectionPipeline(face_detector)
  det_preds = det_pipeline.predict([img])
  det_pipeline.show(det_preds,
                    "/mnt/Lab/results/owlpipelines/aws_facedet.html")
